///////////////////////////////////////////////////////////////////////////////
// ketchup/message_processor.hpp
//
#pragma once

#include <boost/mpl/bool.hpp>
#include <boost/mpl/fold.hpp>
#include <boost/mpl/eval_if.hpp>
#include <boost/mpl/placeholders.hpp>

#include "detail/gen_message_processor_base.hpp"
#include "get_message_map.hpp"
#include "gen_pwm_args_type.hpp"
#include "get_message_map.hpp"
#include "handler/message_handlers.hpp"
#include "handler/cracked_handlers.hpp"
#include "handler/cmd_ui_handlers.hpp"

namespace ketchup {

///////////////////////////////////////////////////////////////////////////////
// message_processor
//
template< class Derived, bool HasCrackedHandlers = false >
struct message_processor :
	detail::gen_message_processor_base< Derived, boost::mpl::bool_<HasCrackedHandlers> >::type,
	handler::message_handlers<Derived>,
	handler::cracked_handlers<Derived>,
	handler::cmd_ui_handlers<Derived>
{
protected:
	typedef typename gen_pwm_args_reference<Derived>::type pwm_args_ref;

	///////////////////////////////////////////////////////////////////////////////
	// process_window_message_to
	//
	template< class Wnd >
	static bool process_window_message_to(Wnd& wnd, pwm_args_ref xs, DWORD dwMsgMapID = 0)
	{
		return wnd.ProcessWindowMessage(get_hWnd<Derived><Derived>(xs), get_uMsg<Derived>(xs),
			get_wParam<Derived>(xs), get_lParam<Derived>(xs), get_lResult<Derived>(xs), dwMsgMapID) == TRUE;
	}

	///////////////////////////////////////////////////////////////////////////////
	// empty_handler
	//
	struct empty_handler
	{
		static bool process(pwm_args_ref xs)
		{
			boost::ignore_unused_variable_warning(xs);
			return false;
		}
	};

	///////////////////////////////////////////////////////////////////////////////
	// chain_msg
	//
	template< class MessageMap, DWORD msgMapID >
	struct chain_msg_alt
	{
		static bool process(pwm_args_ref xs)
		{
			return get_derived<Derived>(xs).process_window_message<MessageMap>(get_hWnd<Derived>(xs), get_uMsg<Derived>(xs),
				get_wParam<Derived>(xs), get_lParam<Derived>(xs), get_lResult<Derived>(xs), msgMapID) == TRUE;
		}
	};

	template< class MessageMap >
	struct chain_msg : chain_msg_alt<MessageMap, 0> { };

	///////////////////////////////////////////////////////////////////////////////
	// chain_msg_map_alt
	//
	template< class BaseMessageProcessor,DWORD msgMapID >
	struct chain_msg_map_alt
	{
		static bool process(pwm_args_ref xs)
		{
			typedef typename get_message_map<BaseMessageProcessor>::type base_message_map;
			return get_derived<Derived>(xs).BaseMessageProcessor::process_window_message<base_message_map>(get_hWnd<Derived>(xs), get_uMsg<Derived>(xs),
				get_wParam<Derived>(xs), get_lParam<Derived>(xs), get_lResult<Derived>(xs), msgMapID) == TRUE;
		}
	};

	template< class BaseMessageProcessor >
	struct chain_msg_map : chain_msg_map_alt<BaseMessageProcessor, 0> { };

	///////////////////////////////////////////////////////////////////////////////
	// chain_msg_map_alt_Atl
	//
	template< class theChainClass, DWORD msgMapID >
	struct chain_msg_map_alt_Atl
	{
		static bool process(pwm_args_ref xs)
		{
			return get_derived<Derived>(xs).theChainClass::ProcessWindowMessage(get_hWnd<Derived>(xs), get_uMsg<Derived>(xs),
				get_wParam<Derived>(xs), get_lParam<Derived>(xs), get_lResult<Derived>(xs), msgMapID) == TRUE;
		}
	};

	template< class theChainClass >
	struct chain_msg_map_Atl : chain_msg_map_alt_Atl<theChainClass, 0> { };

	///////////////////////////////////////////////////////////////////////////////
	// alt_msg_map_member
	//  define CMessageMap& CMainFrame::get_cmdBar() { return &m_wndCmdBar; } and pass to...
	//
	template< CMessageMap& (Derived::*func)() >
	struct chain_msg_map_member
	{
		static bool process(pwm_args_ref xs)
		{
			return process_window_message_to((get_derived<Derived>(xs).*func)(), xs, 0);
		}
	};
	// or define in your CMainFrame if you want chain to non-CMessageMap...
	/*
	struct chain_msg_map_m_wndCmdBar
	{
		static bool process(pwm_args_ref xs)
		{
			return process_window_message_to(get_this(xs).m_just_a_map, xs, 3);
		}
	};
	*/	
	///////////////////////////////////////////////////////////////////////////////
	// chain_msg_map_alt_member
	//
	template< CMessageMap& (Derived::*func)(), DWORD dwMsgMapID >
	struct chain_msg_map_alt_member
	{
		static bool process(pwm_args_ref xs)
		{
			return process_window_message_to((get_derived<Derived>(xs).*func)(), xs, dwMsgMapID);
		}
	};

	///////////////////////////////////////////////////////////////////////////////
	// alt_msg_map
	//
	template< DWORD msgMapID, class MessageMap >
	struct alt_msg_map
	{
		static bool process(pwm_args_ref xs)
		{
			if (get_dwMsgMapID<Derived>(xs) == msgMapID) {
					return get_derived<Derived>(xs).process_window_message<MessageMap>(get_hWnd<Derived>(xs), get_uMsg<Derived>(xs),
						get_wParam<Derived>(xs), get_lParam<Derived>(xs), get_lResult<Derived>(xs), get_dwMsgMapID<Derived>(xs)) == TRUE;
			}
			else
				return false; // pass to other alternates
		}
	};

	///////////////////////////////////////////////////////////////////////////////
	// forward_notifications
	//
	struct forward_notifications
	{
		static bool process(pwm_args_ref xs)
		{
			get_bHandled<Derived>(xs) = TRUE;
			get_lResult<Derived>(xs) = get_derived<Derived>(xs).ForwardNotifications(get_uMsg<Derived>(xs),
				get_wParam<Derived>(xs), get_lParam<Derived>(xs), get_bHandled<Derived>(xs));
			if(get_bHandled<Derived>(xs))
				return true;
		}
	};

	///////////////////////////////////////////////////////////////////////////////
	// default_reflection_handler
	//
	struct default_reflection_handler
	{
		static bool process(pwm_args_ref xs)
		{
			return Derived::DefaultReflectionHandler(get_hWnd<Derived>(xs), get_uMsg<Derived>(xs),
				get_wParam<Derived>(xs), get_lParam<Derived>(xs), get_lResult<Derived>(xs)) == TRUE;
		}
	};

	///////////////////////////////////////////////////////////////////////////////
	// chain_client_commands
	//
	struct chain_client_commands
	{
		static bool process(pwm_args_ref xs)
		{
			if (get_uMsg<Derived>(xs) == WM_COMMAND)
			{
				HWND hWnd = get_derived<Derived>(xs).m_hWndClient;
				if (hWnd != NULL)
					::SendMessage(hWnd, get_uMsg<Derived>(xs), get_wParam<Derived>(xs), get_lParam<Derived>(xs));
				return false;
			}

			return false;
		}
	};

	///////////////////////////////////////////////////////////////////////////////
	// chain_mdi_child_commands
	//
	struct chain_mdi_child_commands
	{
		static bool process(pwm_args_ref xs)
		{
			if (get_uMsg<Derived>(xs) == WM_COMMAND)
			{
				HWND hWndChild = get_derived<Derived>(xs).MDIGetActive();
				if (hWndChild != NULL)
					::SendMessage(hWndChild, get_uMsg<Derived>(xs), get_wParam<Derived>(xs), get_lParam<Derived>(xs));
				return false;
			}
			return false;
		}
	};

}; // struct message_processor

} // namespace ketchup

///////////////////////////////////////////////////////////////////////////////
// KETCHUP_CHAIN_MSG_MAP for ATL's BEGIN_MSG_MAP
// 
#define KETCHUP_CHAIN_MSG_MAP(MessageMap) {\
	if (process_window_message<MessageMap>(hWnd, uMsg, wParam, lParam, lResult, dwMsgMapID))\
		return TRUE;\
}
